-----------------------
----LOCAL VARIABLES----
-----------------------


local ADDON_NAME, namespace = ...
local functions = namespace.functions;
local variables = namespace.variables;
local leaderWindow = CreateFrame("Frame", "LPTLootRollLeader", nil, "BasicFrameTemplateWithInset");
local loadFrame = CreateFrame("FRAME");
local historyIndex = 1;


-----------------------
--FUNCTION DEFINITIONS-
-----------------------


--Function for toggeling of all buttons in a frame.
local function toggleButtons(frame, enabled)
	if enabled then
		frame.mainSpeccButton:Enable();
		frame.offSpeccButton:Enable();
		frame.mogButton:Enable();
		frame.otherButton:Enable();
	else
		frame.mainSpeccButton:Disable();
		frame.offSpeccButton:Disable();
		frame.mogButton:Disable();
		frame.otherButton:Disable();
	end
end

--Function for doing a full copy of a table.
local function deepcopy(orig)
    local orig_type = type(orig);
    local copy;
	
    if orig_type == 'table' then
        copy = {};
        for orig_key, orig_value in next, orig, nil do
            copy[deepcopy(orig_key)] = deepcopy(orig_value);
        end
        setmetatable(copy, deepcopy(getmetatable(orig)));
    else -- number, string, boolean, etc
        copy = orig;
    end
	
    return copy;
end

--Function for clearing a list and a frame.
local function clearList(frame)
	--Clear list views to prepare for new item.
	frame:Clear();
	
	--Reset scrollbar values. 
	frame:GetParent().scrollBar:SetMinMaxValues(0,0);
	frame:GetParent().scrollBar:SetValue(select(1, frame:GetParent().scrollBar:GetMinMaxValues()));
end

--Function for clearing all lists and tables.
local function clearMessageFrames()
	clearList(leaderWindow.mainFrame.messageFrame);
	clearList(leaderWindow.offFrame.messageFrame);
	clearList(leaderWindow.mogFrame.messageFrame);
	clearList(leaderWindow.otherFrame.messageFrame);
end

--Update the names of the lists to reflect their values.
local function updateListNames()
	leaderWindow.mainFrame.title:SetText("Main (Max: " .. llrSettings.mainRoll .. ")");
	leaderWindow.offFrame.title:SetText("Off (Max: " .. llrSettings.offRoll .. ")");
	leaderWindow.mogFrame.title:SetText("Mog (Max: " .. llrSettings.mogRoll .. ")");
end

--Reverse array function 
local function reverseArray(arr)
	local i, j = 1, #arr

	while i < j do
		arr[i], arr[j] = arr[j], arr[i]
		i = i + 1
		j = j - 1
	end
end

--Add array to frame function.
local function addArrayToFrame(frame, array)
	for i,v in ipairs(array) do 
		frame:AddMessage("(" .. array[i][2] .. ") " .. array[i][1], array[i][3], array[i][4], array[i][5]);
	end
end

--Function for updating the message frames.
local function updateMessageframe(frame, array)
	frame:Clear();
	if array == nil then return end;
	local arrLength = # array;
	
	if arrLength > 7 then 
		frame:GetParent().scrollBar:SetMinMaxValues(0, arrLength-7);
	else
		frame:GetParent().scrollBar:SetMinMaxValues(0,0);
	end

	frame:SetMaxLines(arrLength);
	
	--Sort table based primarly by roll number, secondly by name.
	table.sort(array, function(a,b)
		if (a[2] > b[2]) then
			-- primary sort on roll -> a before b
			return true;
		elseif (a[2] < b[2]) then
			-- primary sort on roll -> b before a
			return false;
		else
			-- primary sort tied, resolve with secondary sort on name
			return a[1] < b[1];
		end
	end);
	
	--Determine how the frames should be scrolled before displaying them.
	if arrLength > 0 then
		local scrollBarScroll = 0;
		local frameScroll = arrLength-7;
		if not (llrSettings.scrollMode) then
			reverseArray(array);
			scrollBarScroll = arrLength-7;
			frameScroll = 0;
		end
		addArrayToFrame(frame, array);
		frame:GetParent().scrollBar:SetValue(scrollBarScroll);
		frame:SetScrollOffset(frameScroll);
	end
end

--Function for adding a roller to an array.
local function addRollerToArray(array, roller)
	--Determine if user has already rolled, and if found then exit function.
	if( # array > 0) then
		for i,v in ipairs(array) do 
			if(v[1] == roller[1]) then
				return;
			end
		end
	end
	
	--Clear the current frame and add the roller to the list.
	table.insert(array, roller);
end

--Function for handling page number display.
local function pageControl()
	if(# variables.rollHistory > 1) then
		leaderWindow.historyButtonFrame.pageNumber:SetText(historyIndex .. " / " .. # variables.rollHistory);
		if(# variables.rollHistory >= historyIndex +1) then 
			leaderWindow.historyButtonFrame.nextHistoryPage:Enable();
		else
			leaderWindow.historyButtonFrame.nextHistoryPage:Disable();
		end
	
		if(0 < historyIndex -1) then 
			leaderWindow.historyButtonFrame.previousHistoryPage:Enable();
		else
			leaderWindow.historyButtonFrame.previousHistoryPage:Disable();
		end
	else
		leaderWindow.historyButtonFrame.pageNumber:SetText("1 / 1");
		leaderWindow.historyButtonFrame.nextHistoryPage:Disable();
		leaderWindow.historyButtonFrame.previousHistoryPage:Disable();
	end
end

--Function for handling history list displays.
local function scrollHistory(bool)
	--Check if we're trying to scroll up or down.
	if bool then
		if(# variables.rollHistory >= historyIndex + 1) then 
			historyIndex = historyIndex + 1;
		end
	else
		if(0 < historyIndex - 1) then 
			historyIndex = historyIndex - 1;
		end
	end
	pageControl();
	
	--If historyIndex is 0 that means we are at current item and should display that. Otherwise display item at place historyIndex.
	local id = variables.rollHistory[historyIndex].item:match("item:(%d+):");
	functions.setItemButtonData(leaderWindow.itemButton, variables.rollHistory[historyIndex].item);
	
	updateMessageframe(leaderWindow.mainFrame.messageFrame, variables.rollHistory[historyIndex].mainList);
	updateMessageframe(leaderWindow.offFrame.messageFrame, variables.rollHistory[historyIndex].offList);
	updateMessageframe(leaderWindow.mogFrame.messageFrame, variables.rollHistory[historyIndex].mogList);
	updateMessageframe(leaderWindow.otherFrame.messageFrame, variables.rollHistory[historyIndex].otherList);
end

--Roll list factory.
local function startFrame(frame, config, xValue, yValue, title, length)
	frame.width  = length;
	frame.height = 130;
	frame:SetSize(frame.width, frame.height);
	frame:SetPoint("LEFT", config, "LEFT", xValue, yValue);
	frame:EnableMouse(true);
	frame:EnableMouseWheel(true);
	
	frame.title = frame:CreateFontString(nil,"OVERLAY");
	frame.title:SetFontObject("GameFontHighLight");
	frame.title:SetPoint("TOP", frame, "TOP", 0, 1);
	frame.title:SetText(title);
	
	--ScrollingMessageFrame
	local messageFrame = CreateFrame("ScrollingMessageFrame", nil, frame);
	messageFrame:SetPoint("TOPLEFT", 0, -17);
	messageFrame:SetPoint("BOTTOMRIGHT", 0, 38);
	messageFrame:SetFontObject("GameFontNormalSmall");
	messageFrame:SetTextColor(1, 1, 1, 1);
	messageFrame:SetJustifyH("LEFT");
	messageFrame:SetHyperlinksEnabled(true);
	messageFrame:SetFading(false);
	messageFrame:SetMaxLines(300);
	frame.messageFrame = messageFrame;
	
	local tex = messageFrame:CreateTexture(nil,"BACKGROUND");
	tex:SetAllPoints();
	tex:SetColorTexture(0, 0, 0);
	
	local backDrop = CreateFrame("Frame", nil, messageFrame);
	backDrop:SetPoint("TOPLEFT", -5, 6);
	backDrop:SetPoint("BOTTOMRIGHT", 2, -6);

	backDrop:SetBackdrop( 
	{ 
		bgFile = "Interface/Tooltips/UI-Tooltip-Background", 
		edgeFile = "Interface/Tooltips/UI-Tooltip-Border", 
		tile = true, tileSize = 16, edgeSize = 16, 
		insets = { left = 4, right = 4, top = 4, bottom = 4 }
	});										
	backDrop:SetBackdropColor(0,0,0,0);
	
	---------------
	-- Scroll bar--
	---------------
	
	local scrollBar = CreateFrame("Slider", nil, frame, "UIPanelScrollBarTemplate");
	scrollBar:SetPoint("RIGHT", frame, "RIGHT", 4, 11);
	scrollBar:SetSize(30, frame.height - 85);
	scrollBar:SetMinMaxValues(0, 0);
	scrollBar:SetValueStep(1);
	scrollBar:SetObeyStepOnDrag(true);
	scrollBar.scrollStep = 1;
	frame.scrollBar = scrollBar;

	scrollBar:SetScript("OnValueChanged", function(self, value)
		messageFrame:SetScrollOffset(select(2, scrollBar:GetMinMaxValues()) - value);
	end);
	
	scrollBar:SetValue(0);

	frame:SetScript("OnMouseWheel", function(self, delta)
		local cur_val = scrollBar:GetValue();
		local min_val, max_val = scrollBar:GetMinMaxValues();

		if delta < 0 and cur_val < max_val then
			cur_val = math.min(max_val, cur_val + 1);
			scrollBar:SetValue(cur_val);
		elseif delta > 0 and cur_val > min_val then
			cur_val = math.max(min_val, cur_val - 1);
			scrollBar:SetValue(cur_val);
		end
	end);
end


-----------------------
-----EVENT HANDLING----
-----------------------


loadFrame:RegisterEvent("ADDON_LOADED"); 
loadFrame:RegisterEvent("PLAYER_LOGOUT"); 

loadFrame:SetScript("OnEvent", function(self, event, addonName, ...)
	if event == "ADDON_LOADED" and addonName == "LPT Loot Roll" then
		leaderWindow:SetPoint(llrSettings.raidLeadWindowPos.point, nil, llrSettings.raidLeadWindowPos.relativePoint, llrSettings.raidLeadWindowPos.xOfs, llrSettings.raidLeadWindowPos.yOfs);
		functions.controllEventListener();
		loadFrame:UnregisterEvent("ADDON_LOADED");
	elseif event == "PLAYER_LOGOUT" then
		llrSettings.raidLeadWindowPos.point, _, llrSettings.raidLeadWindowPos.relativePoint, llrSettings.raidLeadWindowPos.xOfs, llrSettings.raidLeadWindowPos.yOfs = leaderWindow:GetPoint();
	end
end);

leaderWindow:SetScript("OnEvent", function(self,event,...)
	local message = ...;
	local author, rollResult, rollMin, rollMax = string.match(message, "(.+) rolls (%d+) %((%d+)-(%d+)%)");
	
	if author then
		local _,playerClass = UnitClass(author);
		local color = RAID_CLASS_COLORS[playerClass];			
		
		--Logic to determine the roll and to then fill in the appropiate list with the roll.
		local rolled = tonumber(rollResult);
		local rolledMin = tonumber(rollMin);
		local rolledMax = tonumber(rollMax);
		if not (rolled == nil or rolledMax == nil or rolledMin ~= 1) then
			local roller = {author, rolled, color.r, color.g, color.b};
			if(rolledMax == llrSettings.mainRoll) then
				addRollerToArray(variables.rollHistory[1].mainList, roller);
			elseif (rolledMax == llrSettings.offRoll) then
				addRollerToArray(variables.rollHistory[1].offList, roller);
			elseif (rolledMax == llrSettings.mogRoll) then
				addRollerToArray(variables.rollHistory[1].mogList, roller);
			else
				--Check that rolledMax is less than 10 000 (to match editbox of other roll).
				if (rolledMax < 10000) then
					roller[2] = rolledMax;
					addRollerToArray(variables.rollHistory[1].otherList, roller);
				end
			end
			if(historyIndex == 1) then
				updateMessageframe(leaderWindow.mainFrame.messageFrame, variables.rollHistory[1].mainList);
				updateMessageframe(leaderWindow.offFrame.messageFrame, variables.rollHistory[1].offList);
				updateMessageframe(leaderWindow.mogFrame.messageFrame, variables.rollHistory[1].mogList);
				updateMessageframe(leaderWindow.otherFrame.messageFrame, variables.rollHistory[1].otherList);
			end
		end
	end
end);


-----------------------
-----LEADER WINDOW-----
-----------------------


leaderWindow:SetSize(340,235);
leaderWindow:SetMovable(true);
leaderWindow:EnableMouse(true);
leaderWindow:RegisterForDrag("LeftButton");
leaderWindow:SetClampedToScreen(true);
leaderWindow:SetClampRectInsets(0, 0, 0, 0)
leaderWindow:SetScript("OnDragStart", function(self) 
	leaderWindow:StartMoving();
end);
leaderWindow:SetScript("OnDragStop", function(self) 
	leaderWindow:StopMovingOrSizing() 
end);
leaderWindow:SetScript("OnShow", function(self) 
	functions.updateLeaderScale();
	leaderWindow:RegisterEvent("CHAT_MSG_SYSTEM");
end);
leaderWindow:SetScript("OnHide", function(self) 
	if (variables.addonLoadedd) then 
		functions.controllEventListener() 
	end;
end);
leaderWindow:SetScript("OnMouseWheel", function(self, delta)
	if # variables.rollHistory == 0 then return end;
	if (delta < 0) and # variables.rollHistory >= historyIndex + 1 then
		scrollHistory(true);
	elseif (delta > 0) and 0 <= historyIndex - 1 then
		scrollHistory(false);
	end
end);

leaderWindow.title = leaderWindow:CreateFontString(nil, "OVERLAY");
leaderWindow.title:SetFontObject("GameFontHighLight");
leaderWindow.title:SetPoint("LEFT", leaderWindow.TitleBg, "LEFT", 5, -1);
leaderWindow.title:SetText("LLR Leader");
leaderWindow:Hide();

--Setup the raid leader item preview.
leaderWindow.itemButton = CreateFrame("Button", nil, leaderWindow, "GameMenuButtonTemplate");
leaderWindow.itemButton:SetPoint("TOPLEFT", leaderWindow, "TOPLEFT", 10, -42);
leaderWindow.itemButton:SetSize(60,60);
leaderWindow.itemButton:SetScript("OnLeave", function(self) GameTooltip:Hide() ResetCursor() end);

leaderWindow.itemButton.itemTexture = leaderWindow.itemButton:CreateTexture(nil,"OVERLAY");
leaderWindow.itemButton.itemTexture:SetPoint("TOPLEFT", 0, 0);
leaderWindow.itemButton.itemTexture:SetPoint("BOTTOMRIGHT", 0, 0);

--Initialize raid leader's roll buttons.
leaderWindow.rollFrame = CreateFrame("Frame","RollButtons", leaderWindow);
leaderWindow.rollFrame:SetSize(60,120);
leaderWindow.rollFrame:SetPoint("LEFT", leaderWindow, "LEFT", 30, 10);

leaderWindow.rollFrame.mainSpeccButton = CreateFrame("Button", nil, leaderWindow.rollFrame, "GameMenuButtonTemplate");
leaderWindow.rollFrame.offSpeccButton = CreateFrame("Button", nil, leaderWindow.rollFrame, "GameMenuButtonTemplate");
leaderWindow.rollFrame.mogButton = CreateFrame("Button", nil, leaderWindow.rollFrame, "GameMenuButtonTemplate");
leaderWindow.rollFrame.otherButton = CreateFrame("Button", nil, leaderWindow.rollFrame, "GameMenuButtonTemplate");

--Start instance of raid leader buttons.
functions.startButton(leaderWindow.rollFrame.mainSpeccButton, leaderWindow.rollFrame, "CENTER", "TOP", -20, -73, "Main", 60, 30, function(self) RandomRoll(1,llrSettings.mainRoll) toggleButtons(leaderWindow.rollFrame, false)  end);
functions.startButton(leaderWindow.rollFrame.offSpeccButton, leaderWindow.rollFrame, "CENTER", "TOP", -20, -103, "Off", 60, 30, function(self) RandomRoll(1,llrSettings.offRoll) toggleButtons(leaderWindow.rollFrame, false) end);
functions.startButton(leaderWindow.rollFrame.mogButton, leaderWindow.rollFrame, "CENTER", "TOP", -20, -133, "Mog", 60, 30, function(self) RandomRoll(1,llrSettings.mogRoll) toggleButtons(leaderWindow.rollFrame, false) end);
functions.startButton(leaderWindow.rollFrame.otherButton, leaderWindow.rollFrame, "CENTER", "TOP", -20, -163, "Other", 60, 30, function(self) StaticPopup_Show("llrManualRoll") end);

--Initalize history buttons
leaderWindow.historyButtonFrame = CreateFrame("Frame","HistoryButtons", leaderWindow);
leaderWindow.historyButtonFrame:SetSize(20,15);
leaderWindow.historyButtonFrame:SetPoint("TOPRIGHT", leaderWindow, "TOPRIGHT", -25, -10);

leaderWindow.historyButtonFrame.pageNumber = leaderWindow:CreateFontString(nil,"OVERLAY");
leaderWindow.historyButtonFrame.pageNumber:SetFontObject("GameFontHighLight");
leaderWindow.historyButtonFrame.pageNumber:SetPoint("RIGHT", leaderWindow.historyButtonFrame, "RIGHT", -50, 6);

leaderWindow.historyButtonFrame.nextHistoryPage = CreateFrame("Button", nil, leaderWindow.historyButtonFrame, "GameMenuButtonTemplate");
leaderWindow.historyButtonFrame.previousHistoryPage = CreateFrame("Button", nil, leaderWindow.historyButtonFrame, "GameMenuButtonTemplate");

leaderWindow.historyButtonFrame.nextHistoryPage:Disable();
leaderWindow.historyButtonFrame.previousHistoryPage:Disable();

--Start instance of history buttons
functions.startButton(leaderWindow.historyButtonFrame.nextHistoryPage, leaderWindow.historyButtonFrame, "CENTER", "TOP", 0, 0, ">", 25, 22, function(self) scrollHistory(true) end);
functions.startButton(leaderWindow.historyButtonFrame.previousHistoryPage, leaderWindow.historyButtonFrame, "CENTER", "TOP", -22, 0, "<", 25, 22, function(self) scrollHistory(false) end);

--Initialize the frames for the rolled lists.
leaderWindow.mainFrame  = CreateFrame("Frame", "MainList", leaderWindow);
leaderWindow.offFrame   = CreateFrame("Frame", "OffList", leaderWindow);
leaderWindow.mogFrame   = CreateFrame("Frame", "MogList", leaderWindow);
leaderWindow.otherFrame = CreateFrame("Frame", "OtherList", leaderWindow);

--Start the instance of the rolled list frames.
startFrame(leaderWindow.mainFrame, leaderWindow, 77, 21, "Main", 120);
startFrame(leaderWindow.offFrame, leaderWindow, 77, -77, "Off", 120);
startFrame(leaderWindow.mogFrame, leaderWindow, 207, 21, "Mog", 120);
startFrame(leaderWindow.otherFrame, leaderWindow, 207, -77, "Other (Max: 9999)", 120);


-----------------------
---GLOBAL FUNCTIONS----
-----------------------

--Function for updating scale of the leader window.
function functions.updateLeaderScale()
	leaderWindow:SetScale(variables.screenScale*llrSettings.leaderScale);
end

--Function to disable leader roll buttons.
function functions.disableLeader()
	toggleButtons(leaderWindow.rollFrame, false);
end

--Function to determine if the leader window is displayed.
function functions.isLeaderShown()
	return leaderWindow:IsShown();
end

--Function for resetting all roll and item history.
function functions.resetRolls()
	clearMessageFrames();
	
	--for i=1, #variables.rollHistory do 
	--	variables.rollHistory[i]=nil 
	--end;
	
	table.wipe(variables.rollHistory);
	
	pageControl();
end

--Function for resetting leader window.
function functions.resetLeaderWindow()
	leaderWindow:ClearAllPoints();
	leaderWindow:SetPoint("CENTER", UIParent, "CENTER");
end

--Function to display the leader window.
function functions.displayLeader()
	--Toggle leaderWindow UI.
	if not leaderWindow:IsShown() then
		leaderWindow:Show();
	end 
	
	updateListNames();
		
	if llrSettings.saveMode and variables.rollHistory[1] then 
		scrollHistory(false);
	end
end

--Function for starting the leader window upon a new item drop.
function functions.leaderEvent(isEquipable)
	--Setup the item "frame" with tooltip, mog and artifact system implementation.
	functions.setItemButtonData(leaderWindow.itemButton, variables.rollHistory[1].item);
	
	--Clear all roll frames and lists.
	clearMessageFrames();
	historyIndex = 1;
	
	--Prepare history buttons.
	leaderWindow.historyButtonFrame.previousHistoryPage:Disable();
	pageControl();
	
	--Show raid leader UI.
	leaderWindow:Show();
	updateListNames();
	
	--If leader can equip, enable buttons, otherwise disable.
	if isEquipable then
		toggleButtons(leaderWindow.rollFrame, true);
	else
		toggleButtons(leaderWindow.rollFrame, false);
	end
end

--Function for controlling the event listener for rolls.
function functions.controllEventListener()
	if (llrSettings.eventMode) then
		leaderWindow:RegisterEvent("CHAT_MSG_SYSTEM");
	else
		leaderWindow:UnregisterEvent("CHAT_MSG_SYSTEM"); 
	end
end